# vim: set fileencoding=utf-8 :
#
# (C) 2013 Guido Günther <agx@sigxcpu.org>
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, please see
#    <http://www.gnu.org/licenses/>
"""Parts shared between the deb and rpm commands"""

import re
import os
import traceback
from gbp.errors import GbpError
from gbp.deb import DebianPkgPolicy
from gbp.pkg import Archive
from gbp.deb.upstreamsource import DebianAdditionalTarball


class ExitCodes(object):
    ok = 0,
    failed = 1               # All other errors
    no_value = 2             # Value does not exist (gbp config only)
    parse_error = 3          # Failed to parse configuration file
    uscan_up_to_date = 4     # Uscan up to date (import-orig only)


def maybe_debug_raise():
    if 'raise' in os.getenv("GBP_DEBUG", '').split(','):
        raise


def is_download(args):
    """
    >>> is_download(["http://foo.example.com"])
    True
    >>> is_download([])
    False
    >>> is_download(["foo-1.1.orig.tar.gz"])
    False
    """
    if args and re.match("https?://", args[0]):
        return True
    return False


# FIXME: this could become a method of DebianUpstreamSource
def get_component_tarballs(name, version, tarball, components):
    """
    Figure out the paths to the component tarballs based on the main
    tarball.
    """
    tarballs = []
    (_, _, comp_type) = Archive.parse_filename(tarball)
    for component in components:
        cname = DebianPkgPolicy.build_tarball_name(name,
                                                   version,
                                                   comp_type,
                                                   os.path.dirname(tarball),
                                                   component)
        sig = cname + '.asc'
        if not os.path.exists(sig):
            sig = None
        tarballs.append(DebianAdditionalTarball(cname, component, sig=sig))
        if not os.path.exists(cname):
            raise GbpError("Cannot find component tarball %s" % cname)
    return tarballs


def debug_exc(options):
    if options.verbose:
        traceback.print_exc()



# vim: set fileencoding=utf-8 :
#
# (C) 2013 Intel Corporation <markus.lehtonen@linux.intel.com>
# (C) 2016 Guido Günther <agx@sigxcpu.org>
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, please see
#    <http://www.gnu.org/licenses/>
"""Debian-specific upstream sources"""

from gbp.pkg import UpstreamSource
from gbp.deb.policy import DebianPkgPolicy
import gbp.command_wrappers

import os
import shutil
import tempfile


class DebianUpstreamSource(UpstreamSource):
    """Upstream source class for Debian"""
    def __init__(self, name, unpacked=None, sig=None):
        super(DebianUpstreamSource, self).__init__(name,
                                                   unpacked=unpacked,
                                                   sig=sig,
                                                   pkg_policy=DebianPkgPolicy)


class DebianAdditionalTarball(DebianUpstreamSource):
    """Upstream source class for additional tarballs"""
    def __init__(self, name, component, unpacked=None, sig=None):
        self.component = component
        super(DebianAdditionalTarball, self).__init__(name,
                                                      unpacked=unpacked,
                                                      sig=sig)

    def unpack(self, dest, filters):
        """
        Unpack the additional tarball into {dir} naming it
        I{component}.  Apply filters during unpack.

        @param dir: the main tarball dir
        @param filters: filters to apply

        We can't simply use unpack since we need to remove any preexisting dirs and
        name the target directory after the component name.
        """
        olddir = os.path.abspath(os.path.curdir)
        tmpdir = None
        try:
            tmpdir = os.path.abspath(tempfile.mkdtemp(dir=os.path.join(dest, '..')))
            super(DebianAdditionalTarball, self).unpack(tmpdir, filters)

            newdest = os.path.join(dest, self.component)
            if os.path.exists(newdest):
                shutil.rmtree(newdest)
            shutil.move(self.unpacked, newdest)
        finally:
            os.chdir(olddir)
            if tmpdir is not None:
                gbp.command_wrappers.RemoveTree(tmpdir)()



# vim: set fileencoding=utf-8 :
#
# (C) 2021 Andrej Shadura <andrew@shadura.me>
# (C) 2021 Collabora Limited
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, please see
#    <http://www.gnu.org/licenses/>

import os

from tests.component import (ComponentTestBase,
                             ComponentTestGitRepository)
from tests.component.deb.fixtures import RepoFixtures

from nose.tools import ok_

from gbp.scripts.clone import main as clone
from gbp.scripts.setup_gitattributes import main as setup_gitattributes


sorted_gitattrs = [
    '# Added by git-buildpackage to disable .gitattributes found in the upstream tree',
    '* -export-ignore',
    '* -export-subst',
    '* dgit-defuse-attrs',
    '[attr]dgit-defuse-attrs  -text -eol -crlf -ident -filter -working-tree-encoding',
]


dgit_attributes = [
    '*\tdgit-defuse-attrs',
    '[attr]dgit-defuse-attrs\t-text -eol -crlf -ident -filter -working-tree-encoding',
    '# ^ see GITATTRIBUTES in dgit(7) and dgit setup-new-tree in dgit(1)'
]


dgit34_attributes = [
    '*\tdgit-defuse-attrs',
    '[attr]dgit-defuse-attrs\t-text -eol -crlf -ident -filter',
    '# ^ see dgit(7).  To undo, leave a definition of [attr]dgit-defuse-attrs'
]


class TestSetupGitattributes(ComponentTestBase):
    """Test setting up Git attributes"""

    @RepoFixtures.native()
    def test_setup_gitattrs_do_nothing(self, repo):
        """Test that disabling all knows presets works and gives an error"""
        dest = os.path.join(self._tmpdir,
                            'cloned_repo')
        clone(['arg0',
               repo.path, dest])
        cloned = ComponentTestGitRepository(dest)
        self._check_repo_state(cloned, 'master', ['master'])

        attrs_file = os.path.join(dest, '.git', 'info', 'attributes')

        # file shouldn’t exist at this point yet
        self.assertFalse(os.path.exists(attrs_file) and os.path.getsize(attrs_file),
                         "%s is non-empty" % attrs_file)

        os.chdir(cloned.path)
        setup_gitattributes(['arg0', '--no-dgit-defuse-attrs'])

        # file shouldn’t exist at this point yet
        self.assertFalse(os.path.exists(attrs_file) and os.path.getsize(attrs_file),
                         "%s is non-empty" % attrs_file)

        self._check_log(-1,
                        "gbp:error: Nothing to do, no settings to apply.")

    @RepoFixtures.native()
    def test_setup_gitattrs(self, repo):
        """Test that setting up Git attributes manually works"""
        dest = os.path.join(self._tmpdir,
                            'cloned_repo')
        clone(['arg0',
               repo.path, dest])
        cloned = ComponentTestGitRepository(dest)
        self._check_repo_state(cloned, 'master', ['master'])

        attrs_file = os.path.join(dest, '.git', 'info', 'attributes')

        # file shouldn’t exist at this point yet
        self.assertFalse(os.path.exists(attrs_file) and os.path.getsize(attrs_file),
                         "%s is non-empty" % attrs_file)

        os.chdir(cloned.path)
        setup_gitattributes(['arg0'])

        ok_(os.path.exists(attrs_file), "%s is missing" % attrs_file)

        with open(attrs_file) as f:
            attrs = sorted(f.read().splitlines())
        self.assertEquals(attrs, sorted_gitattrs)

    @RepoFixtures.native()
    def test_setup_gitattrs_dgit(self, repo):
        """Test that setting up Git attributes manually works even with dgit"""
        dest = os.path.join(self._tmpdir,
                            'cloned_repo')
        clone(['arg0',
               repo.path, dest])
        cloned = ComponentTestGitRepository(dest)
        self._check_repo_state(cloned, 'master', ['master'])

        attrs_file = os.path.join(dest, '.git', 'info', 'attributes')

        # file shouldn’t exist at this point yet
        self.assertFalse(os.path.exists(attrs_file) and os.path.getsize(attrs_file),
                         "%s is non-empty" % attrs_file)

        with open(attrs_file, 'w') as f:
            f.write('\n'.join(dgit_attributes))

        os.chdir(cloned.path)
        setup_gitattributes(['arg0', '--verbose'])

        ok_(os.path.exists(attrs_file), "%s is missing" % attrs_file)

        with open(attrs_file) as f:
            attrs = sorted(f.read().splitlines())

        expected_gitattrs = sorted(sorted_gitattrs + [
            '# Old dgit macro disabled:',
            '# [attr]dgit-defuse-attrs\t-text -eol -crlf -ident -filter -working-tree-encoding',
            '# ^ see GITATTRIBUTES in dgit(7) and dgit setup-new-tree in dgit(1)',
        ])

        self.assertEquals(attrs, expected_gitattrs)

    @RepoFixtures.native()
    def test_setup_gitattrs_dgit34(self, repo):
        """Test that setting up Git attributes manually works even with a very old dgit"""
        dest = os.path.join(self._tmpdir,
                            'cloned_repo')
        clone(['arg0',
               repo.path, dest])
        cloned = ComponentTestGitRepository(dest)
        self._check_repo_state(cloned, 'master', ['master'])

        attrs_file = os.path.join(dest, '.git', 'info', 'attributes')

        # file shouldn’t exist at this point yet
        self.assertFalse(os.path.exists(attrs_file) and os.path.getsize(attrs_file),
                         "%s is non-empty" % attrs_file)

        with open(attrs_file, 'w') as f:
            f.write('\n'.join(dgit34_attributes))

        os.chdir(cloned.path)
        setup_gitattributes(['arg0', '--verbose'])

        ok_(os.path.exists(attrs_file), "%s is missing" % attrs_file)

        with open(attrs_file) as f:
            attrs = sorted(f.read().splitlines())

        expected_gitattrs = sorted(sorted_gitattrs + [
            '# Old dgit macro disabled:',
            '# [attr]dgit-defuse-attrs\t-text -eol -crlf -ident -filter',
            '# ^ see dgit(7).  To undo, leave a definition of [attr]dgit-defuse-attrs',
        ])

        self.assertEquals(attrs, expected_gitattrs)



# vim: set fileencoding=utf-8 :
#
# (C) 2006-2011, 2016 Guido Günther <agx@sigxcpu.org>
# (C) 2021 Andrej Shadura <andrew@shadura.me>
# (C) 2021 Collabora Limited
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, please see
#    <http://www.gnu.org/licenses/>
#
"""Common repository setup functionality."""

import os
import re
import gbp.log

from pathlib import Path


def set_user_name_and_email(repo_user, repo_email, repo):
    if repo_user == 'DEBIAN':
        if os.getenv('DEBFULLNAME'):
            repo.set_user_name(os.getenv('DEBFULLNAME'))

    if repo_email == 'DEBIAN':
        if os.getenv('DEBEMAIL'):
            repo.set_user_email(os.getenv('DEBEMAIL'))


def check_gitattributes(repo, treeish) -> bool:
    """
    Verify the treeish doesn’t contain non-empty .gitattributes files.
    """
    for mode, _type, sha1, size, path in repo.list_tree(treeish, recurse=True, sizes=True):
        if size == 0:
            continue
        if path == b'.gitattributes' or path.endswith(b'/.gitattributes'):
            gbp.log.debug("Found non-empty .gitattributes: %s" % path)
            return False
    return True


dgit_attr_macro_re = re.compile(r'^\[attr\]dgit-defuse-attrs\s')
dgit_attr_macro_defn = '-text -eol -crlf -ident -filter -working-tree-encoding'
attr_glob_defns = {
    'dgit-defuse-attrs',
    '-export-subst',
    '-export-ignore',
}


def is_gitattributes_set_up(repo) -> bool:
    """
    Return True if git attributes have been set up correctly:
        - dgit-defuse-attrs macro exists
        - dgit-defuse-attrs includes attributes we’re interested in
        - dgit-defuse-attrs is enabled for *
        - export-subst and export-ignore are unset for *
    """
    gitattrs = Path(repo.git_dir) / 'info' / 'attributes'
    if not gitattrs.exists():
        return False
    dgit_macro_present = False
    attrs = set()
    for line in gitattrs.read_text().splitlines():
        if dgit_attr_macro_re.match(line):
            gbp.log.debug("Found Git attribute macro: %s" % line)
            dgit_macro_present = line.endswith(dgit_attr_macro_defn)
            continue
        attr = line.split()
        if attr[0] == '*' and len(attr) == 2:
            attrs.add(attr[1])
    gbp.log.debug("Found global Git attributes: %s" % ', '.join(attrs))
    return dgit_macro_present and attrs >= attr_glob_defns


def setup_gitattributes(repo, treeish='HEAD'):
    """
    Setup .git/info/attributes in a way to prevent transformations from interfering
    with packaging, because the working tree files can differ from the Git revision
    history (and from the source packages).

    Similar functionality has been implemented by dgit and git-deborig, so we try
    to stay compatible and re-use the name of the attribute macro. Since dgit doesn’t
    disable export-subst and export-ignore, which may interfere with export-orig, we
    add this on top the same way git-deborig does.
    """
    if is_gitattributes_set_up(repo):
        return
    gbp.log.debug("Configuring Git attributes")
    gitattrs = Path(repo.git_dir) / 'info' / 'attributes'
    new_attributes = []
    if not gitattrs.exists():
        gitattrs.parent.mkdir(exist_ok=True)
    else:
        for line in gitattrs.read_text().splitlines():
            attr = line.split()
            if attr[0] == '*' and len(attr) == 2:
                if attr[1] in attr_glob_defns:
                    continue
            if dgit_attr_macro_re.match(line):
                new_attributes += [
                    "# Old dgit macro disabled:",
                    "# %s" % line,
                ]
                gbp.log.debug("Disabling old dgit macro: '%s'" % line)
            else:
                new_attributes.append(line)
    new_attributes += [
        "# Added by git-buildpackage to disable .gitattributes found in the upstream tree",
        "[attr]dgit-defuse-attrs  %s" % dgit_attr_macro_defn,
    ] + ['* %s' % attr for attr in attr_glob_defns]
    with gitattrs.with_suffix('.new') as newattrs:
        newattrs.write_text('\n'.join(new_attributes))
        newattrs.rename(gitattrs)



# vim: set fileencoding=utf-8 :
#
# (C) 2021 Andrej Shadura <andrew@shadura.me>
# (C) 2021 Collabora Limited
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, please see
#    <http://www.gnu.org/licenses/>
#
"""Setup Git attributes to incapacitate .gitattributes shipped by the upstream"""

import os
import sys
import gbp.log
from gbp.command_wrappers import CommandExecFailed
from gbp.config import GbpOptionParserDebian
from gbp.deb.git import GitRepositoryError, DebianGitRepository
from gbp.errors import GbpError
from gbp.scripts.common import ExitCodes
from gbp.scripts.common.repo_setup import setup_gitattributes


def build_parser(name):
    setting_presets = {'dgit-defuse-attrs'}

    def disable_preset(option, opt, value, parser):
        setting = opt.replace('--no-', '')
        if parser.values.attr_presets and setting in parser.values.attr_presets:
            parser.values.attr_presets.remove(setting)

    def enable_preset(option, opt, value, parser):
        setting = opt.replace('--', '')
        parser.values.attr_presets.add(setting)

    try:
        parser = GbpOptionParserDebian(command=os.path.basename(name), prefix='',
                                       usage='%prog - set up sane Git attributes')
    except GbpError as err:
        gbp.log.err(err)
        return None

    parser.add_option("--verbose", action="store_true", dest="verbose",
                      default=False, help="verbose command execution")
    for preset in setting_presets:
        parser.add_option("--%s" % preset, action="callback", callback=enable_preset,
                          help="Apply %s preset" % preset)
        parser.add_option("--no-%s" % preset, action="callback", callback=disable_preset,
                          help="Do not apply %s preset" % preset)
    parser.add_option("--all", action="store_const", dest="attr_presets",
                      const=setting_presets, default=setting_presets, help="apply all known settings")
    parser.add_config_file_option(option_name="color", dest="color",
                                  type='tristate')
    parser.add_config_file_option(option_name="color-scheme",
                                  dest="color_scheme")

    return parser


def parse_args(argv):
    """Parse the command line arguments
    @return: options and arguments
    """

    parser = build_parser(argv[0])
    if not parser:
        return None, None

    (options, args) = parser.parse_args(argv[1:])
    gbp.log.setup(options.color, options.verbose, options.color_scheme)
    return options, args


def main(argv):
    repo = None

    (options, args) = parse_args(argv)
    if not options:
        return ExitCodes.parse_error

    try:
        try:
            repo = DebianGitRepository('.')
        except GitRepositoryError:
            raise GbpError("%s is not a git repository" % (os.path.abspath('.')))

        if not options.attr_presets:
            raise GbpError("Nothing to do, no settings to apply.")
        setup_gitattributes(repo)
    except (GitRepositoryError, GbpError, CommandExecFailed) as err:
        if str(err):
            gbp.log.err(err)
    except KeyboardInterrupt:
        gbp.log.err("Interrupted. Aborting.")

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))

# vim:et:ts=4:sw=4:et:sts=4:ai:set list listchars=tab\:»·,trail\:·:



# vim: set fileencoding=utf-8 :
#
# (C) 2012 Guido Günther <agx@sigxcpu.org>
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, please see
#    <http://www.gnu.org/licenses/>
"""Handle checkin and checkout of archives from the pristine-tar branch"""

import re
import os
import gbp.log
from gbp.command_wrappers import Command


class PristineTar(Command):
    """The pristine-tar branch in a git repository"""
    branch = 'pristine-tar'

    def __init__(self, repo):
        self.repo = repo
        super(PristineTar, self).__init__('pristine-tar',
                                          cwd=repo.path,
                                          capture_stderr=True)

    def _has_in_output(self, match):
        """
        Check if pristine_tar has a certain feature enabled.

        @param feature: feature / command option to check
        @type feature: C{str}
        @return: True if feature is supported
        @rtype: C{bool}
        """
        self.call(['--help'], quiet=True)  # There's no --help so we always exit 1
        r = re.compile(match)
        for line in self.stderr.splitlines():
            if r.match(line):
                return True
        return False

    def has_feature_verify(self):
        """Does this pristine-tar support tarball verification"""
        return self._has_in_output(".* pristine-tar .* verify")

    def has_feature_sig(self):
        """Does this pristine-tar support detached upstream signatures"""
        return self._has_in_output(r'.*--signature-file')

    def has_commit(self, archive_regexp):
        """
        Do we have a pristine-tar commit for a package matching I{archive_regexp}.

        @param archive_regexp: archive name to look for (regexp wildcards allowed)
        @type archive_regexp: C{str}
        """
        return True if self.get_commit(archive_regexp)[0] else False

    def _commit_contains_file(self, commit, regexp):
        """Does the given commit contain a file with the given regex"""
        files = self.repo.get_commit_info(commit)['files']
        # CPython wants '+' (which is valid in source package names)
        # escaped but git-grep doesn't so we do it that late:
        cregex = re.compile(regexp.replace('+', '\\+'))
        for _, v in files.items():
            for f in v:
                if cregex.match(f.decode()):
                    return True
        return False

    def get_commit(self, archive_regexp):
        """
        Get the pristine-tar commit of a package matching I{archive_regexp}.
        Checks also whether the commit contains a signature file.

        @param archive_regexp: archive name to look for (regexp wildcards allowed)
        @type archive_regexp: C{str}
        @return: Commit, True if commit contains a signature file
        @rtype: C{tuple} of C{str} and C{bool}
        """
        if not self.repo.has_pristine_tar_branch():
            return None, False

        regex = ('pristine-tar .* %s' % archive_regexp)
        commits = self.repo.grep_log(regex, self.branch, merges=False)
        if commits:
            commit = commits[-1]
            gbp.log.debug("Found pristine-tar commit at '%s'" % commit)
            return commit, self._commit_contains_file(commit, '%s.asc' % archive_regexp)
        return None, False

    def checkout(self, archive, quiet=False, signaturefile=None):
        """
        Checkout an orig archive from pristine-tar branch

        @param archive: the name of the orig archive
        @type archive: C{str}
        """
        args = ['checkout', archive]
        self.run_error = 'Pristine-tar couldn\'t checkout "%s": {stderr_or_reason}' % os.path.basename(archive)
        if signaturefile and self.has_feature_sig():
            args += ['-s', signaturefile]
        self.__call__(args, quiet=quiet)

    def commit(self, archive, upstream, quiet=False, signaturefile=None):
        """
        Commit an archive I{archive} to the pristine tar branch using upstream
        branch ${upstream}.

        @param archive: the archive to commit
        @type archive: C{str}
        @param upstream: the upstream branch to diff against
        @type upstream: C{str}
        """
        args = ['commit', archive, upstream]
        self.run_error = ("Couldn't commit to '%s' with upstream '%s': {stderr_or_reason}" %
                          (self.branch, upstream))
        if signaturefile and self.has_feature_sig():
            args += ['-s', signaturefile]
        self.__call__(args, quiet=quiet)

    def verify(self, archive, quiet=False):
        """Verify an archive's I{archive} checksum using to the pristine tar branch"""

        self.run_error = 'Pristine-tar couldn\'t verify "%s": {stderr_or_reason}' % os.path.basename(archive)
        self.__call__(['verify', archive], quiet=quiet)
